var gltfTutorial = gltfTutorial || {};

define([
    'Utils'
], function (
    Utils
){
    (function(gltfTutorial) {
        'use strict';

        var optional = gltfTutorial.Utils.optional;

        /**
         * Constructor for a GL context. A GlContext contains the actual 
         * WebGL rendering context, and offers functions for creating the
         * OpenGL representations of glTF objects, based on the data that
         * was obtained from the glTF: An OpenGL program can be created
         * from the vertex- and fragment shader source code of the glTF
         * program. An OpenGL buffer can be created from a glTF bufferView
         * and the binary data of its buffer. An OpenGL texture can be
         * created from a glTF texture and the ImageData of the glTF image 
         * that the glTF texture refers to. 
         *
         * @param {Object} canvas The canvas for which the GL context
         * will be created
         */
        function GlContext(canvas) {

            this.gl = undefined;
            try {
                this.gl = canvas.getContext("experimental-webgl");
            } catch (e) {
                alert(e.message);
            }
            if (!this.gl) {
                alert("Could not initialize WebGL");
                return;
            }
            this.glViewport = {
                width : canvas.width,
                height : canvas.height
            };
            this.gl.clearColor(0.0, 0.0, 0.0, 1.0);

            // TODO This should come from the glTF!
            this.gl.enable(this.gl.DEPTH_TEST);

            /**
             * Create an OpenGL shader with the given type from the given shader
             * source code and return the OpenGL ID of the shader.
             *
             * If the shader can not be compiled, then the return value is undefined.
             *
             * @param {String} shaderSource The shader source code
             * @param {Integer} shaderType The shader type
             * @return {Integer} The OpenGL shader ID
             */
            function createGlShader(shaderSource, shaderType) {
                var gl = this.gl;
                var glShader = gl.createShader(shaderType);
                gl.shaderSource(glShader, shaderSource);
                gl.compileShader(glShader);
                if (!gl.getShaderParameter(glShader, gl.COMPILE_STATUS)) {
                    alert("Error: " + gl.getShaderInfoLog(glShader));
                    return;
                }
                return glShader;
            }

            /**
             * Create an OpenGL program from the given vertex- and fragment
             * shader, and return the OpenGL ID of the program.
             *
             * If the program can not be linked, then the return value is undefined.
             *
             * @param {Integer} glVertexShader The OpenGL ID of the vertex shader
             * @param {Integer} glFragmentShader The OpenGL ID of the fragment shader
             * @return {Integer} The OpenGL program ID
             */
            function createAndLinkGlProgram(glVertexShader, glFragmentShader) {
                var gl = this.gl;
                var glProgram = gl.createProgram();
                gl.attachShader(glProgram, glVertexShader);
                gl.attachShader(glProgram, glFragmentShader);
                gl.linkProgram(glProgram);
                if (!gl.getProgramParameter(glProgram, gl.LINK_STATUS)) {
                    alert("Error: " + gl.getProgramInfoLog(glProgram));
                    return;
                }
                return glProgram;
            }

            /**
             * Create an OpenGL program from the given vertex- and fragment shader
             * source code
             *
             * @param vertexShaderSource {String} The vertex shader source code
             * @param fragmentShaderSource {String} The fragment shader source code
             * @return The OpenGL program ID
             */
            GlContext.prototype.createGlProgram = function(vertexShaderSource, fragmentShaderSource) {
                var gl = this.gl;
                var glVertexShader = createGlShader.call(this, vertexShaderSource, gl.VERTEX_SHADER);
                var glFragmentShader = createGlShader.call(this, fragmentShaderSource, gl.FRAGMENT_SHADER);
                var glProgram = createAndLinkGlProgram.call(this, glVertexShader, glFragmentShader);
                return glProgram;
            };

            /**
             * Create an OpenGL buffer for the given glTF bufferView
             *
             * @param {Object} bufferView The glTF bufferView
             * @param {arraybuffer} bufferData The data of the glTF buffer
             * that is referred to by the given bufferView
             * @return {Integer} The OpenGL buffer ID
             */
            GlContext.prototype.createGlBuffer = function(bufferView, bufferData) {
                var gl = this.gl;
                var glBuffer = gl.createBuffer();
                gl.bindBuffer(bufferView.target, glBuffer);
                var dataView = new DataView(bufferData, bufferView.byteOffset, bufferView.byteLength);
                gl.bufferData(bufferView.target, dataView, gl.STATIC_DRAW);
                return glBuffer;
            };

            /**
             * Create an OpenGL texture for the given glTF texture
             *
             * @param {Object} texture The glTF texture
             * @param {ImageData} imageData The data of the image
             * that is referred to by the given texture
             * @return {Integer} The OpenGL texture ID
             */
            GlContext.prototype.createGlTexture = function(texture, imageData) {
                var gl = this.gl;
                var glTexture = gl.createTexture();

                var target = optional(texture.target, gl.TEXTURE_2D);
                gl.bindTexture(target, glTexture);

                var internalFormat = optional(texture.internalFormat, gl.RGBA);
                var format = optional(texture.format, gl.RGBA);
                var type = optional(texture.type, gl.UNSIGNED_BYTE);

                gl.texImage2D(target, 0, internalFormat, format, type, imageData);

                var sampler = texture.sampler;
                var magFilter = optional(sampler.magFilter, gl.LINEAR);
                var minFilter = optional(sampler.minFilter, gl.NEAREST_MIPMAP_LINEAR);
                var wrapS = optional(sampler.wrapS, gl.REPEAT);
                var wrapT = optional(sampler.wrapT, gl.REPEAT);

                gl.texParameteri(target, gl.TEXTURE_MAG_FILTER, magFilter);
                gl.texParameteri(target, gl.TEXTURE_MIN_FILTER, minFilter);
                gl.texParameteri(target, gl.TEXTURE_WRAP_S, wrapS);
                gl.texParameteri(target, gl.TEXTURE_WRAP_T, wrapT);
                gl.generateMipmap(target);

                return glTexture;
            };
        };

        gltfTutorial.GlContext = GlContext;

    })(gltfTutorial);
});

